<?php

namespace Arsenii\WebSockets;

use Arsenii\WebSockets\Facades\Emitter;
use Arsenii\WebSockets\Facades\Looper;
use Arsenii\WebSockets\Connection;

use Arsenii\WebSockets\Log;

use Workerman\Worker;

final class Server
{
     /**
     * Version.
     *
     * @var string
     */
    const VERSION = '3.0.1';

    /**
     * Status starting.
     *
     * @var int
     */
    const STATUS_STARTING = 1;

    /**
     * Status running.
     *
     * @var int
     */
    const STATUS_RUNNING = 2;

    /**
     * Status shutdown.
     *
     * @var int
     */
    const STATUS_SHUTDOWN = 4;
    
    /**
     * The backlog for WebSockets Server.
     *
     * @var string
     */
    protected $backlog;
    
    /**
     * The host address for WebSockets Server.
     *
     * @var string
     */
    protected $hostAddress;
    
    /**
     * The host port for WebSockets Server.
     *
     * @var string
     */
    protected $hostPort;
    
    /**
     * The host path for WebSockets Server.
     *
     * @var string
     */
    protected $hostPath;
    
    /**
     * The host protocol for WebSockets Server.
     *
     * @var string
     */
    protected $hostProtocol;

    /**
     * The host private key for WebSockets Server.
     *
     * @var string
     */
    protected $hostPrivateKey;
    
    /**
     * The host certificate for WebSockets Server.
     *
     * @var string
     */
    protected $hostCert;
    
    
    /**
     * The host passphrase for WebSockets Server.
     *
     * @var string
     */
    protected $hostPassphrase;

    /**
     * The array of connections.
     *
     * @var array
     */
    protected $connections = [];

    /**
     * Current status.
     *
     * @var int
     */
    protected static $_status = self::STATUS_STARTING;
    

    /**
     * Create a new WebSockets Server instance.
     *
     * @param  string|array|null    $hostAddress
     * @param  string|null          $hostPort
     * @param  string|null          $hostPath
     * @param  string|null          $hostProtocol
     * @param  string|null          $hostDefaultServantName
     * @return void
     */
    public function __construct($hostAddress = null, $hostPort = null, $hostPath = null, $hostProtocol = null){
        
        Log::info('init server', Log::LEVEL_DEBUG);

        // Allow only CLI Mode
        if (php_sapi_name() != "cli") {

            exit("only run in command line mode \n");
        }
        
        // Import configs from config_path('websockets.php')
        $default = config('websockets');

        // hostAddress argument can be passed like array
        if( is_array( $hostAddress ) ){

            $default = array_merge( $default, $hostAddress );

        }else{

            if(! is_null( $hostAddress ) )
                $default['host'] = $hostAddress;

            if(! is_null( $hostPort ) )
                $default['port'] = $hostPort;

            if(! is_null( $hostPath ) )
                $default['path'] = $hostPath;

            if(! is_null( $hostProtocol ) )
                $default['protocol'] = $hostProtocol;

        }

        // Set Server variable from default array formed
        $this->setVars( $default );

        $this->backlog = md5(
            ( $this->hostProtocol != "auto" ? $this->hostProtocol : '(wss?)' ).
            '://'.
            $this->hostAddress.
            ':'.
            $this->hostPort.
            ( substr( $this->hostPath, 0, 1 ) == '/' ? '' : '/' ).
            $this->hostPath
        );

    }

    /**
     * Assing vars to a newly WebSockets Server instance.
     *
     * @param  array|null  $vars
     * @return void
     */
    protected function setVars($vars = []){
        
        if( isset( $vars['host'] ) ){

            $this->hostAddress      = $vars['host'];
        }

        if( isset( $vars['port'] ) ){

            $this->hostPort         = $vars['port'];
        }

        if( isset( $vars['path'] ) ){

            $this->hostPath         = $vars['path'];
        }

        if( isset( $vars['protocol'] ) ){

            $this->hostProtocol     = $vars['protocol'];
        }

        if( isset( $vars['ssl'] ) ){

            if( isset( $vars['ssl']['local_private_key'] ) ){
                
                $this->hostPrivateKey   = $vars['ssl']['local_private_key'];
            }
            
            if( isset( $vars['ssl']['local_certificate'] ) ){
                
                $this->hostCert         = $vars['ssl']['local_certificate'];
            }
            
            if( isset( $vars['ssl']['passphrase'] ) ){
                
                $this->hostPassphrase   = $vars['ssl']['passphrase'];
            }
        }

    }

    /**
     * Start WebSockets Server instance.
     *
     * @return void
     */
    public function start(){

        global $argv;

        Log::info('open server', Log::LEVEL_DEBUG);

        $address                = $this->hostAddress .':'. $this->hostPort .( substr( $this->hostPath, 0, 1 ) == '/' ? '' : '/' ). $this->hostPath;
        $context_options        = null;

        // If protocol not ws enable ssl
        if( $this->hostProtocol != 'ws' ){

            $context_options    = [
                    'ssl' => [
                        'local_cert'    => $this->hostCert,
                        'local_pk'      => $this->hostPrivateKey,                        
                        'verify_peer'   => !1,
                    ]
                ];
            
            // If `passphrase` was added push it in context
            if(! empty( $this->hostPassphrase ) ){

                $context_options['ssl']['passphrase'] = $this->hostPassphrase;
            }

            $worker = new Worker("websocket://$address", $context_options);
            $worker->transport = 'ssl';

        } else {

            $worker = new Worker("websocket://$address");
        }


        $worker->count = 1;

        $worker->onConnect = function($target) { 
            
            $this->makeConection($target);
        };

        $worker->onMessage = function($target, $data){

            Log::info("message-received [ $data ]");

            Emitter::dispatch( 'message-received', $this->connections[ $target->id ], $data );

        };
        
        $worker->onClose = function($target){
            
            if( isset( $this->connections[ $target->id ] ) ) {

                $this->connections[ $target->id ]->close();
            }
        };
        
        $argv = [
            'websockets:start',
            'start'
        ];

        $worker->onWorkerStart = function() use($address) { 

            Log::info("Server [ ". ( $this->hostProtocol != "auto" ? $this->hostProtocol : '(wss?)' ) .'://'. $address ." ] are started");
    
            $this->clearBacklog();
    
            // Load all Listeners from base_path('routes/websockets.php')
            Emitter::addListeners();

            Looper::start();
        };
        
        Worker::runAll();  

    }

    /**
     * Accept new connection.
     *
     * @param  string  $target
     * @return void
     */
    protected function makeConection($target){

        // Assign New Connection to server connections array
        $this->connections[ $target->id ] = new Connection($target, $this);

        Log::info( 'new Conn -> '. $target->id, Log::LEVEL_DEBUG);

    }

    /**
     * Unset connection.
     *
     * @param  string  $connuid
     * @return bool
     */
    public function unsetConnection($connuid){

        // Check Connection uniqid presence
        if ( isset( $this->connections[ $connuid ] ) ) {

            unset( $this->connections[ $connuid ] );
            
        }

        return true;
    }

    public function clearBacklog(){
        
        if(! is_dir( __DIR__."/Backlog" ) )
            mkdir( __DIR__."/Backlog" );

        touch( __DIR__."/Backlog/{$this->backlog}.log" );

        unlink( __DIR__."/Backlog/{$this->backlog}.log" );

        if( is_dir( sys_get_temp_dir() ."/{$this->backlog}" ) ){
            
            array_map('unlink', glob(sys_get_temp_dir() ."/{$this->backlog}/*.*"));
            rmdir( sys_get_temp_dir() ."/{$this->backlog}" );
        }
    }

    public function checkBacklog(){

        if(! is_dir( __DIR__."/Backlog" ) )
            mkdir( __DIR__."/Backlog" );

        touch( __DIR__."/Backlog/{$this->backlog}.log" );

        $lines  = file( __DIR__."/Backlog/{$this->backlog}.log", FILE_IGNORE_NEW_LINES | FILE_SKIP_EMPTY_LINES );

        foreach( $lines as $index => $line ){
            
            if( substr($line, -10) === ' [checked]' )
                continue;

            $message = file_get_contents( $line );

            Emitter::dispatch( 'backlog', null, $message );

            $lines[ $index ] = $line .' [checked]';
            
            unlink( $line );
        }

        file_put_contents( __DIR__."/Backlog/{$this->backlog}.log", implode("\n", $lines) );
        
    }

    public function putBacklog( string $data = '' ){

        if(! is_dir( __DIR__."/Backlog" ) )
            mkdir( __DIR__."/Backlog" );
        
        if(! is_dir( sys_get_temp_dir() ."/{$this->backlog}" ) ){
            
            mkdir( sys_get_temp_dir() ."/{$this->backlog}" );
        }

        $tmp = null;

        while(
                ($tmp = sys_get_temp_dir() ."/{$this->backlog}/". uniqid($this->backlog) .'.log' ) != null
            &&  file_exists( $tmp )
        );

        file_put_contents( $tmp, $data );

        $backlog = fopen(__DIR__."/Backlog/{$this->backlog}.log", 'a');

        fwrite($backlog, "\n".$tmp);

        fclose($backlog);

    }

    static public function instance( $address = null, $port = null, $path = null, $protocol = null ){

        return new self( $address, $port, $path, $protocol );
    }

    public function getConnections(){

        return $this->connections;
    }
}
